@@ -126,6 +126,82 @@ IOObjectGetClass(
     return( io_object_get_class( object, className ));
 }
 
+
+CFStringRef 
+IOObjectCopyClass(io_object_t object)
+{
+	io_name_t my_name;
+	CFStringRef my_str = NULL;
+	
+	// if there's no argument, no point going on.  Return NULL.
+	if (object == NULL) {
+		return my_str;
+	}
+	
+	io_object_get_class( object, &my_name );	
+	my_str = CFStringCreateWithCString (kCFAllocatorDefault, my_name, kCFStringEncodingUTF8);
+	
+	return my_str;
+}
+
+CFStringRef 
+IOObjectCopySuperclassForClass(CFStringRef classname)
+{
+	io_name_t my_name, orig_name;
+	CFStringRef my_str = NULL;
+	char * my_cstr;
+	kern_return_t kr; 
+	
+	mach_port_t masterPort = __IOGetDefaultMasterPort();
+	
+	// if there's no argument, no point going on.  Return NULL.
+	if (classname == NULL) {
+		return my_str;
+	}
+	
+	my_cstr = malloc(sizeof(char) * 128);
+	CFStringGetCString (classname, my_cstr, 128, kCFStringEncodingUTF8);
+		
+	strncpy(orig_name, my_cstr, sizeof(io_name_t));
+	
+	kr = io_object_get_superclass(masterPort, &orig_name, &my_name);
+	if (kr == kIOReturnSuccess) {
+		my_str = CFStringCreateWithCString (kCFAllocatorDefault, my_name, kCFStringEncodingUTF8);
+	}
+	free(my_cstr);
+	
+	return my_str;
+}
+
+CFStringRef 
+IOObjectCopyBundleIdentifierForClass(CFStringRef classname)
+{
+	io_name_t my_name, orig_name;
+	CFStringRef my_str = NULL;
+	char * my_cstr;
+	kern_return_t kr; 
+	
+	mach_port_t masterPort = __IOGetDefaultMasterPort();
+	
+	// if there's no argument, no point going on.  Return NULL.
+	if (classname == NULL) {
+		return my_str;
+	}
+	
+	my_cstr = malloc(sizeof(char) * 128);
+	CFStringGetCString (classname, my_cstr, 128, kCFStringEncodingUTF8);
+		
+	strncpy(orig_name, my_cstr, sizeof(io_name_t));
+	
+	kr = io_object_get_bundle_identifier(masterPort, &orig_name, &my_name);
+	if (kr == kIOReturnSuccess) {
+		my_str = CFStringCreateWithCString (kCFAllocatorDefault, my_name, kCFStringEncodingUTF8);
+	}
+	free(my_cstr);
+	
+	return my_str;
+}
+
 boolean_t
 IOObjectConformsTo(
 	io_object_t	object,
@@ -237,6 +313,7 @@ IOServiceGetMatchingServices(
 	return( kIOReturnBadArgument);
 
     data = IOCFSerialize( matching, kNilOptions );
+    CFRelease( matching );
     if( !data)
 	return( kIOReturnUnsupported );
 
@@ -255,7 +332,6 @@ IOServiceGetMatchingServices(
     }
 
     CFRelease( data );
-    CFRelease( matching );
 
     return( kr );
 }
@@ -314,6 +390,7 @@ IOServiceAddNotification(
 	return( kIOReturnBadArgument);
 
     data = IOCFSerialize( matching, kNilOptions );
+    CFRelease( matching );
     if( !data)
 	return( kIOReturnUnsupported );
 
@@ -334,7 +411,6 @@ IOServiceAddNotification(
     }
 
     CFRelease( data );
-    CFRelease( matching );
 
     return( kr );
 }
@@ -357,6 +433,7 @@ IOServiceAddMatchingNotification(
 	return( kIOReturnBadArgument);
 
     data = IOCFSerialize( matching, kNilOptions );
+    CFRelease( matching );
     if( !data)
 	return( kIOReturnUnsupported );
 
@@ -384,7 +461,6 @@ IOServiceAddMatchingNotification(
     }
 
     CFRelease( data );
-    CFRelease( matching );
 
     return( kr );
 }
@@ -628,6 +704,64 @@ OSGetNotificationFromMessage(
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * * */
 
+#if __ppc__
+static void swapOSNotification(
+       OSNotificationHeader * header,
+       unsigned int leftOver )
+{
+    // Do some conditional byte-swapping if we're running a ppc binary, 
+    // under Rosetta on x86 HW.
+    // We expect this function to be called ONLY when running in Rosetta, not
+    // when running on native ppc HW.
+    
+    header->size = OSSwapInt32(header->size);
+    header->type = OSSwapInt32(header->type);
+    
+    int i;
+    for (i = 0; i < sizeof(header->reference) / sizeof(header->reference[0]); i++) {
+           header->reference[i] = OSSwapInt32(header->reference[i]);
+    }
+    
+    switch (header->type) {
+        case kIOServicePublishNotificationType:
+        case kIOServiceMatchedNotificationType:
+        case kIOServiceTerminatedNotificationType:
+               // only need to swap the reference field
+               break;
+        case kIOAsyncCompletionNotificationType:
+        {
+            IOAsyncCompletionContent *asyncHdr = 
+                    (IOAsyncCompletionContent *)(header + 1);
+            leftOver = (leftOver - sizeof(*asyncHdr)) / sizeof(void *);
+            
+            asyncHdr->result = OSSwapInt32(asyncHdr->result);
+            for (i = 0; i < leftOver; i++) {
+                asyncHdr->args[i] = OSSwapInt32(asyncHdr->args[i]);
+            }
+            break;
+        }
+        case kIOServiceMessageNotificationType:
+        {
+            IOServiceInterestContent *interestHdr = 
+                    (IOServiceInterestContent *)(header + 1);
+            leftOver = (leftOver - sizeof(*interestHdr) + 
+                    sizeof(interestHdr->messageArgument)) / sizeof(void *);
+            
+            interestHdr->messageType = OSSwapInt32(interestHdr->messageType);
+            for (i = 0; i < leftOver; i++) {
+                interestHdr->messageArgument[i] = 
+                        OSSwapInt32(interestHdr->messageArgument[i]);
+            }
+            break;
+        }
+        default:
+               printf("unknown OSNotificationHeader type %d\n", header->type);
+               break;
+    }
+}
+#endif
+
+
 void
 IODispatchCalloutFromMessage(void *cfPort, mach_msg_header_t *msg, void *info)
 {
@@ -665,6 +799,15 @@ IODispatchCalloutFromCFMessage(CFMachPortRef port, void *_msg, CFIndex size, voi
 
     leftOver = msg->msgh_size - (((vm_address_t) (header + 1)) - ((vm_address_t) msg));
 
+#if __ppc__
+    // If this is ppc code running under Rosetta over an x86 kernel,
+    // we may need to do some byte swapping. The magic cookie in the
+    // inline asm makes Rosetta ignore this branch command.
+    asm volatile ("b L_SkipSwap_label \n .long 0x14400004");
+    swapOSNotification(header, leftOver);
+    asm volatile ("L_SkipSwap_label:");
+#endif
+
     // remote port is the notification (an iterator_t) that fired
     notifier = msg->msgh_remote_port;
 
@@ -926,7 +1069,7 @@ IOConnectSetNotificationPort(
 
 kern_return_t
 IOConnectMapMemory(
-	mach_port_t	connect,
+	io_connect_t	connect,
 	unsigned int	memoryType,
 	task_port_t	intoTask,
 	vm_address_t *	atAddress,
@@ -939,7 +1082,7 @@ IOConnectMapMemory(
 
 kern_return_t
 IOConnectUnmapMemory(
-	mach_port_t	connect,
+	io_connect_t	connect,
 	unsigned int	memoryType,
 	task_port_t	intoTask,
 	vm_address_t	atAddress )
@@ -1800,14 +1943,14 @@ IORegisterClient(
 
 kern_return_t
 IORegistryDisposeEnumerator(
-	mach_port_t	enumerator )
+	io_enumerator_t	enumerator )
 {
     return( IOObjectRelease( enumerator ));
 }
 
 kern_return_t
 IOMapMemory(
-	mach_port_t	connect,
+	io_connect_t	connect,
 	unsigned int	memoryType,
 	task_port_t	intoTask,
 	vm_address_t *	atAddress,
